/* gc-info -- summarize information from the Penn State gene-conversion package
*
* syntax "gc-info all.gc", "gc-info all.gc tabs", or "gc-info all.gc few"
*
* where all.gc is generated by the PSU package. Add "tabs" too produce
* tab-delimited output (suitiable for a Microsoft Word table), or "few" to
* produce a tab-delimited table with fewer columns.
*
* It must be run in a gene-cluster directory (with seq.d and exons.d
* subdirectories) that has been processed by the PSU gene-conversion package.
*/

#include "lib.h"
#include <dirent.h> // for reading directories

#define BUF_LEN 10000	// bounds length of line in all.gc
#define MAF_LEN 10000	// bounds length of line in a self-alignment MAF file
#define MAX_EVENT 2000	// bounds number of unique conversion events per species
#define HUGE 10000000	// bounds length of an analyzed gene cluster

// which fields of all.sh are used?
int pick_field[] = {1, 12, 13, 14, 15, 25};
char field[6][100];
char *primate[] = {
"human",
"chimp",
"chimpnzee",
"gorilla",
"orangutan",
"orang",
"gibbon",
"colobus_monkey",
"baboon",
"vervet",
"rhesus",
"marmoset",
"dusky_titi",
"owl_monkey",
"spider_monkey",
"squirrel_monkey",
"lemur",
"black_lemur",
"mouse_lemur",
"galago"
};

// 0-1 arrays showing positions involved in conversion or duplication
char c[HUGE], d[HUGE];

char cur_spe[1000],	// name of current species
     *spe[1000];	// set of species having an identified conversion
int  max_end,		// last converted position in current species
     nuniq,		// number of unique conversions for current species
     nspe,		// number of species with an identified conversion
     npri = sizeof(primate)/sizeof(primate[0]),
     apri,		// actual number of primate species
     few,		// create fewer columns for a Microsoft Word document
     tabs;		// create table for a Microsoft Word document
float
     tot_pct_dup,	// total primate %conv dups
     tot_pct_gene,	// total primate %conv genes
     tot_pct_CDS;	// total primate %conv CDS,

int is_primate(char *species) {
	int i;

	for (i = 0; i < npri; ++i)
		if (same_string(species, primate[i]))
			return 1;
	return 0;
}

// count bases in self-alignments (duplications) for current species
int get_dup() {
	FILE *fp;
	int i, beg, len, end, big_end = 0, dup_bp;
	char filename[100];
	char buf[MAF_LEN];

	//sprintf(filename, "ortho.d/self.d/%s.maf", cur_spe);
	sprintf(filename, "%s.d/%s.remove_repeats.maf", cur_spe, cur_spe);
	fp = ckopen(filename, "r");
	while (fgets(buf, MAF_LEN, fp))
		if (buf[0] == 's') {
			if (sscanf(buf, "%*s %*s %d %d", &beg, &len) != 2)
				fatalf("gag: %s", buf);
			end = beg + len;
			big_end = MAX(big_end, end);
			if (big_end >= HUGE)
				fatal("cluster is too big");
			for (i = beg+1; i <= end; ++i) // base-0 open
				d[i] = 1;
		}
	fclose(fp);
	for (i = dup_bp = 0; i <= big_end; ++i)
		if (d[i]) {
			++dup_bp;
			d[i] = 0;
		}
	return dup_bp;
}

// extract specified fields of a line from all.gc
void get_fields(char *b) {
	char *f;
	int i, j, nfield = sizeof(pick_field)/sizeof(pick_field[0]);

	f = strtok(b, "\t\n");
	for (i = j = 0; j < nfield && f != NULL; f = strtok(NULL, "\t\n"), ++i)
		if (i == pick_field[j])
			strcpy(field[j++], f);
	if (j != nfield)
		fatalf("Too many fields specified.");
}

// list the species for which no conversions have been detected
void no_conv() {
	DIR *dp;
	struct dirent *ep;
	int i, comma;
     
	dp = opendir ("seq.d/");
	if (dp == NULL)
		fatal("Couldn't open the directory seq.d");

	comma = 0;
	while ((ep = readdir(dp)) != NULL)
		if (ep->d_name[0] != '.') {
			for (i = 0;
			     i < nspe && !same_string(ep->d_name, spe[i]); ++i)
				;
			if (i == nspe) {
				if (comma)
					putchar(',');
				else
					printf("no conversions:");
				comma = 1;
				printf(" %s", ep->d_name);
			}
		}
	(void) closedir (dp);
	if (comma)
		putchar('\n');
}

float pct(int i, int j) {
	if (i == 0)
		return 0.0;
	else if (j == 0)
		fatalf("cannot handle non-zero divided by 0");
	return 100.0*(float)i/(float)j;
}

// print the line summarizing conversions for the current species
void report() {
	FILE *ep;
	int i, beg, end, conv, ngene, conv_gene, cds_bp, conv_cds_bp, dup_bp,
	  hit;
	char filename[1000], buf[100], *u;
	float p, q;

	if (cur_spe[0] == 0)
		return;
	if ((tabs || few) && !is_primate(cur_spe))
		return;
	
	spe[nspe++] = copy_string(cur_spe);
	if (is_primate(cur_spe))
		++apri;
	else if (tabs || few)
		fprintf(stderr, "%s is not a primate\n", cur_spe);
	if ((tabs || few) && strchr(cur_spe, '_')) {
		strcpy(buf, cur_spe);
		u = strchr(buf, '_');
		*u = ' ';
		printf("%s\t", buf);
	} else
		printf("%s\t", cur_spe);
	if (!tabs && !few && strlen(cur_spe) < 8)
		putchar('\t');

	// how much is duplicated and how much was converted?
	for (i = conv = 0; i <= max_end; ++i)
		conv += c[i];
	dup_bp = get_dup();
	p = pct(conv,dup_bp);
	if (few)
		printf("%d\t%1.1f\t%d\t%1.1f\t", nuniq,
		  (float)conv/(float)nuniq, dup_bp, p);
	else if (tabs)
		printf("%d\t%1.1f\t%d\t%d\t%1.1f\t", nuniq,
		  (float)conv/(float)nuniq, dup_bp, conv, p);
	else
		printf("%5d\t%7.1f\t%6d\t%6d\t%6.1f\t", nuniq,
		  (float)conv/(float)nuniq, dup_bp, conv, p);
	if (is_primate(cur_spe))
		tot_pct_dup += p;
	// count genes, converted gene, coding bp, and converted coding bp
	sprintf(filename, "exons.d/%s.exons", cur_spe);
	ep = ckopen(filename, "r");
	ngene = conv_gene = cds_bp = conv_cds_bp = hit = 0;
	while (fgets(buf, 100, ep))
		if (buf[0] == '>' || buf[0] == '<') {	// start a new gene
			conv_gene += hit;
			hit = 0;
			++ngene;
		} else if (sscanf(buf, "%d %d", &beg, &end) == 2) {	// exon
			cds_bp += (end - beg + 1);
			for (i = beg; i <= end; ++i) // base-1, closed interval
				if (c[i]) {
					++conv_cds_bp;
					hit = 1;
				}
		}
	conv_gene += hit;
	fclose(ep);
	p = pct(conv_gene, ngene);
	q = pct(conv_cds_bp, cds_bp);
	if (few)
		printf("%d\t%1.1f\t%d\t%1.1f\n", ngene, p, cds_bp, q);
	else if (tabs)
		printf("%d\t%d\t%1.1f\t%d\t%d\t%1.1f\n",
		  ngene, conv_gene, p, cds_bp, conv_cds_bp, q);
	else
		printf("%5d\t%5d\t%6.1f\t%7d\t%6d\t%6.1f\n",
		   ngene, conv_gene, p, cds_bp, conv_cds_bp, q);
	if (is_primate(cur_spe)) {
		tot_pct_gene += p;
		tot_pct_CDS += q;
	}
}

int main(int argc, char **argv) {
	FILE *fp;
	char buf[BUF_LEN];
	int beg1, end1, beg2, end2, i, event, uniq_event[MAX_EVENT]; 

	if (argc == 3) {
		if (same_string(argv[2], "tabs"))
			tabs = 1;
		else if (same_string(argv[2], "few"))
			few = 1;
		else
			fatal("Second arg must be \"tabs\" or \"few\"");
	} else if (argc != 2)
		fatalf("syntax: gc-info all.gc [new] [tabs]");
	fp = ckopen("all.gc", "r");

	if (few) {
		printf("species\t#conv\tave len\tdup bp\t%%conv\t");
		printf("#gene\t%%conv\t#CDS bp\t%%conv\n");
	} else if (tabs) {
		printf("species\t#conv\tave len\tdup bp\tconv\t%%conv\t");
		printf("#gene\tconv\t %%conv\t#CDS bp\tconv\t %%conv\n");
	} else {
		printf("species\t\t#conv\tave_len\tdup_bp\t  conv\t %%conv\t");
		printf("#gene\t conv\t %%conv\t#CDS_bp\t  conv\t %%conv\n");
	}

	// discard tree
	if (fgets(buf, BUF_LEN, fp) == NULL || buf[0] != '(')
		fatalf("tree? %s", buf);
	while (fgets(buf, BUF_LEN, fp)) {	// scan all.gc
		get_fields(buf);
		if (!same_string(cur_spe, field[0])) {	// new species
			report();
			for (i = 0; i <= max_end; ++i)
				c[i] = 0;
			nuniq = max_end = 0;
			strcpy(cur_spe, field[0]);
		}
		beg1 = atoi(field[1]);
		end1 = atoi(field[2]);
		beg2 = atoi(field[3]);
		end2 = atoi(field[4]);
		max_end = MAX(max_end, MAX(end1,end2));
		if (max_end >= HUGE)
			fatal("cluster is too big");
		// mark positions involved in this observation of a conversion
		for (i = beg1; i <= end1; ++i)
			c[i] = 1;
		for (i = beg2; i <= end2; ++i)
			c[i] = 1;
		event = atoi(field[5]);
		// seen this conversion event before?
		for (i = 0; i < nuniq && event != uniq_event[i]; ++i)
			;
		if (nuniq >= MAX_EVENT)
			fatalf("%s has too many conversion events", cur_spe);
		if (i == nuniq)
			uniq_event[nuniq++] = event;
	}
	report();	// don't forget the last species

	// list species with no detected conversions
	no_conv();
	fprintf(stderr, "\nfor primates:\n");
	fprintf(stderr,
	 "  %4.1f%% of duplicated positions in (source or target of) a conversion\n",
	  tot_pct_dup/(float)apri);
	fprintf(stderr,
	 "  %4.1f%% of protein-coding positions in a conversion\n",
	  tot_pct_CDS/(float)apri);
	fprintf(stderr,
	 "  %4.1f%% of genes have at least one position in a conversion\n",
	  tot_pct_gene/(float)apri);

	return 0;
}
